(* ILang.t -> TreeAsm.t *)

open Shared
open Global
open ILang
open TreeAsm

let onreg = GlobalReg.result

let is_small x = x <= 16
let is_16bit i = -0x8000 <= i && i < 0x8000
let is_6bit i = -0x20 <= i && i < 0x20

let g_floor x =
  let a = new_str () in
  let b = new_str () in
  let c = new_str () in
  let d = new_str () in
  let e = new_str () in
  Let(a, Imm(F 8388608.),
      Ret(IfI(Lt,x,0,
              Let(c, Op2(FAdd,x,a),
                  Let(d, Op2(FSub,c,a),
                      Ret(FIf(Eq,x,d,Ret(Mov d),
                              Let(e, Imm(F 1.0), Ret(Op2(FSub,d,e))))))),
              Let(b, Op2(FSub,x,a), Ret(Op2(FAdd,b,a))))))

let rec g e =
  let insert_let x k =
    match x with
    | Var x -> k (Id.to_string x)
    | Unit -> k reg_dummy
    | Int 0 | Float 0. -> k reg_zero
    | _ ->
        let y = new_str () in
        let i = match x with
          | Int i -> I i
          | Float a -> F a
          | Label l -> L l
          | _ -> assert false
        in
        Let(y, Imm i, k y)
  in
  let addi x i = assert (is_16bit i);
    insert_let x
      (fun x -> Ret(Op1i(AddI, x, I i)))
  in
  let op1 o x =
    insert_let x
      (fun x -> Ret(Op1(o, x)))
  in
  let op2 o x y =
    insert_let x
      (fun x -> insert_let y
         (fun y -> Ret(Op2(o, x, y))))
  in
  let if_typ x =
    match x with
    | Unit -> assert false
    | Int _ | Label _ -> Type.Tint
    | Float _ -> Type.Tfloat
    | Var x -> x.Id.typ
  in
  let ifi cmp x i e1 e2 =
    if is_6bit i then insert_let x
      (fun x -> Ret(IfI(cmp, x, i, g e1, g e2)))
    else insert_let x
      (fun x -> insert_let (Int i)
         (fun y -> Ret(If(cmp, x, y, g e1, g e2))))
  in
  let ifv cmp x_ y_ e1 e2 =
    insert_let x_
      (fun x -> insert_let y_
         (fun y -> if if_typ x_ = Type.Tfloat then
            Ret(FIf(cmp, x, y, g e1, g e2))
          else
            Ret(If(cmp, x, y, g e1, g e2))))
  in
  match e with
  | V Unit -> Ret Nop
  | V(Int i) -> Ret(Imm(I i))
  | V(Float a) -> Ret(Imm(F a))
  | V(Label l) -> Ret(Imm(L l))
  | V(Var x) -> Ret(Mov(Id.to_string x))
  | Op(Label ".+", [x; Int i]) when is_16bit i -> addi x i
  | Op(Label ".+", [Int i; x]) when is_16bit i -> addi x i
  | Op(Label ".-", [x; Int i]) when is_16bit (-i) -> addi x (-i)
  | Op(Label ".+", [x; y]) -> op2 Add x y
  | Op(Label ".-", [x; y]) -> op2 Sub x y
  | Op(Label ".*", [x; y]) -> g (Op(Label ".mul", [x; y]))
  | Op(Label "./", [x; y]) -> g (Op(Label ".div", [x; y]))
  | Op(Label ".lsl", [x; Int i]) ->
      insert_let x
        (fun x -> Ret(Op1i(Sll, x, I(i land 0x1f))))
  | Op(Label ".asr", [x; Int i]) ->
      insert_let x
        (fun x -> Ret(Op1i(Srr, x, I(i land 0x1f))))
  | Op(Label ".+.", [x; y]) -> op2 FAdd x y
  | Op(Label ".*.", [Float (-1.); x]) | Op(Label ".*.", [x; Float (-1.)])
  | Op(Label ".-.", [Float 0.; x]) -> insert_let x
        (fun x -> Ret(Op2(Xnor,x,reg_maxint)))
  | Op(Label ".-.", [x; y]) -> op2 FSub x y
  | Op(Label ".*.", [x; y]) -> op2 FMul x y
  | Op(Label "./.", [Float 1.; y]) -> op1 FInv y
  | Op(Label "./.", [x; y]) ->
      insert_let x
        (fun x -> insert_let y
           (fun y -> let z = new_str () in
            Let(z, Op1(FInv, y), Ret(Op2(FMul, x, z)))))
  | Op(Label ".sqrt", [x]) -> op1 FSqrt x
  | Op(Label ".floor", [x]) -> insert_let x g_floor
  | Op(Label ".fabs", [x]) -> insert_let x
        (fun x -> Ret(Op2(And,x,reg_maxint)))
  | Op(Label ".xor", [x; y]) -> op2 Xor x y
  | Op(Label ".xnor", [x; y]) -> op2 Xnor x y
  | Op(Label ".&", [x; y]) -> op2 And x y
  | Op(Label ".or", [x; y]) -> op2 Or x y
  | Op(Label ".not", [x]) -> g (Op(Label ".-", [Int 1; x]))
  | Op(Label ".print_char", [x]) -> insert_let x
        (fun x -> Ret(Op1(Wr, x)))
  | Op(Label ".read_char", _) -> Ret Rd
  | Op(Label ".create_array", [Int i; _]) when i <= 0 -> Ret Nop
  | Op(Label ".create_array", [Int i; y]) when is_small i ->
      insert_let y
        (fun y -> let rec iter j =
           if j >= i then let z = new_str () in
           Let(z, Mov reg_hp, Let(reg_hp, Op1i(AddI, reg_hp, I i), Ret(Mov z)))
           else Let(reg_dummy, St(y, reg_hp, I j), iter (j+1))
         in
         iter 0)
  | Op(Label ".create_array", [x; y]) -> op2 Array x y
  | Op(Label ".<-", [Label l; Int i; z]) when Hashtbl.mem onreg l ->
      let oft = Hashtbl.find onreg l in
      insert_let z (fun z -> Ret(St(z, reg_zero, L regs_gl.(oft + i))))
  | Op(Label ".->", [Label l; Int i]) when Hashtbl.mem onreg l ->
      let oft = Hashtbl.find onreg l in
      Ret(Mov regs_gl.(oft + i))
  | Op(Label ".<-", Label l :: _) when Hashtbl.mem onreg l -> assert false
  | Op(Label ".->", Label l :: _) when Hashtbl.mem onreg l -> assert false
  | Op(Label ".<-", [Label l; Int i; z])
  | Op(Label ".t<-", [Label l; Int i; z]) ->
      let li = l ^ "." ^ string_of_int i in
      insert_let z
        (fun z -> Ret(St(z, reg_zero, L li)))
  | Op(Label ".<-", [Label l; y; z]) ->
      insert_let y
        (fun y -> insert_let z
           (fun z -> Ret(St(z, y, L(l ^ ".0")))))
  | Op(Label ".<-", [x; Int i; z]) ->
      insert_let x
        (fun x -> insert_let z
           (fun z -> Ret(St(z, x, I i))))
  | Op(Label ".<-", [x; y; z]) ->
      let w = new_str () in
      insert_let x
        (fun x -> insert_let y
           (fun y -> insert_let z
              (fun z -> Let(w, Op2(Add, x, y), Ret(St(z, w, I 0))))))
  | Op(Label ".->", [Label l; Int i]) ->
      let li = l ^ "." ^ string_of_int i in
      Ret(Op1i(Ld, reg_zero, L li))
  | Op(Label ".->", [Label l; y])
  | Op(Label ".t->", [Label l; y]) ->
      insert_let y
        (fun y -> Ret(Op1i(Ld, y, L(l ^ ".0"))))
  | Op(Label ".->", [x; Int i])
  | Op(Label ".t->", [x; Int i]) ->
      insert_let x
        (fun x -> Ret(Op1i(Ld, x, I i)))
  | Op(Label ".->", [x; y]) ->
      let w = new_str () in
      insert_let x
        (fun x -> insert_let y
           (fun y -> Let(w, Op2(Add, x, y), Ret(Op1i(Ld, w, I 0)))))
  | Op(Label ".float_of_bits", [x]) | Op(Label ".bits_of_float", [x]) ->
      g (V x)
  | Op(Label ".tuple", xs) ->
      let rec iter i xs =
        match xs with
        | [] -> let y = new_str () in
          Let(y, Mov reg_hp, Let(reg_hp, Op1i(AddI, reg_hp, I i), Ret(Mov y)))
        | x::xs -> insert_let x
              (fun x -> Let(reg_dummy, St(x, reg_hp, I i), iter (i+1) xs))
      in
      iter 0 xs
  | Op(Label l, xs) ->
      let rec iter xs ys =
        match xs with
        | [] -> Ret(Call(l, List.rev ys))
        | x::xs -> insert_let x (fun x -> iter xs (x::ys))
      in
      iter xs []
  | ILang.If(ILang.Eq,x,Int i,e1,e2)
  | ILang.If(ILang.Eq,Int i,x,e1,e2) -> ifi Eq x i e1 e2
  | ILang.If(LE,x,Int i,e1,e2) -> ifi Lt x (i+1) e1 e2
  | ILang.If(LE,Int i,x,e1,e2) -> ifi Lt x i e2 e1
  | ILang.If(ILang.Eq,x,y,e1,e2) -> ifv Eq x y e1 e2
  | ILang.If(LE,x,y,e1,e2) -> ifv Lt y x e2 e1
  | ILang.Let(x,e1,e2) ->
      if x.Id.typ = Type.Tunit then concat (g e1) reg_dummy (g e2)
      else concat (g e1) (Id.to_string x) (g e2)
  | _ -> assert false

let f e =
  let funs = !gl_funs in
  gl_funs := [];
  { name = "main";
    args = [];
    body = g e;
    stksiz = 0 } ::
  List.map
    (fun { Global.name = name; Global.args = ys; Global.body = body } ->
      if List.length ys > Array.length regs then
        failwith (name ^ " has too large arity");
      let live = ILang.fv body in
      let xs = List.map
          (fun y -> if S.mem y live then Id.to_string y else reg_dummy)
          ys
      in
      { name = name;
        args = xs;
        body = g body;
        stksiz = 0 })
    funs
